//    WINMOCK -- Windows Sockets Non-Network Version for testing
//      sockets apps off-line
// WinMinS.C -- minimal sockets layer
// Copyright (C) Joseph Hlavaty, 1995.  All rights reserved.
#include <windows.h>
#include <winsock.h>
#include "winmock.h"

// This routine is called by our main() routine, which is the
//   actual entrypoint for the library from Windows...
BOOL APIENTRY LIBMAIN( HANDLE hInstance, WORD wDataSeg, WORD cbHeap )
{
   wmInitializeSocketList() ;
   wmInitializeTaskList() ;

   return( TRUE ) ;
}

// will cause any future WinSock API calls using the given socket to fail
//  this routine frees the socket for use
int APIENTRY closesocket ( SOCKET s )
{
   int rc ;
   HTASK ht = GetCurrentTask() ;
   PMYTASK pthistask ;
   PMYSOCKET pthissocket = wmGetSocketInfo( s ) ;

   DebugBreak() ;

   // if this app isn't on our task list, then they
   //    haven't called WSAStartup() or have already called WSACleanup()
   pthistask = wmGetTaskInfo( ht ) ;
   if ( NULL == pthistask )
   {
     return( SOCKET_ERROR ) ;
   }

   if ( NULL == pthissocket )
   {
      pthistask->mylasterror = WSAENOTSOCK ;
      return( SOCKET_ERROR ) ;
   }

   // if socket doesn't belong to this task, then return error
   if ( ht != pthissocket->mytask )
   {
      pthistask->mylasterror = WSAENOTSOCK ;
      return SOCKET_ERROR ;
   }

   // mark socket as no longer in use...
   wmClearSocketStruct( pthissocket ) ;

// permit the user to have a WINMOCK.INI file which can have the following format
// [closesocket]
// rc=<error code>
// a note about error processing:  if a [lasterror] section exists in winmock.ini
//   then we check it for a key name of this api's name.  If found, then we
//   assign this value to our "last error" value and abort the operation
//   Otherwise, we get the user rc from the <apiname> section, keyname "rc"
   rc = GetPrivateProfileString( "lasterror", "closesocket",
          "",
          szErrorString, ERRORSTRINGSIZE,
          "WINMOCK.INI" ) ;
   if ( 0 == rc )
   {
      rc = GetPrivateProfileString( "closesocket", "rc",
          "ZERO",
          szErrorString, ERRORSTRINGSIZE,
          "WINMOCK.INI" ) ;
      rc = wmConvertErrorString( szErrorString ) ;
   }
   else
   {
      rc = wmConvertErrorString( szErrorString ) ;
      pthistask->mylasterror = rc ;
      return( SOCKET_ERROR ) ;
   }

   return( rc ) ;
}

// establish a network connection -- a virtual circuit
int APIENTRY connect (SOCKET s, LPSOCKADDR name, int namelen )
{
   HTASK ht = GetCurrentTask() ;
   PMYTASK pthistask ;
   PMYSOCKET pthissocket = wmGetSocketInfo( s ) ;
   int rc ;

   DebugBreak() ;

   // if this app isn't on our task list, then they
   //    haven't called WSAStartup() or have already called WSACleanup()
   pthistask = wmGetTaskInfo( ht ) ;
   if ( NULL == pthistask )
   {
     return( SOCKET_ERROR ) ;
   }

   if ( NULL == pthissocket )
   {
      pthistask->mylasterror = WSAENOTSOCK ;
      return( SOCKET_ERROR ) ;
   }

   // if socket doesn't belong to this task, then return error
   if ( ht != pthissocket->mytask )
   {
      pthistask->mylasterror = WSAENOTSOCK ;
      return SOCKET_ERROR ;
   }

   // if socket is already connected, then error
   if ( TRUE == pthissocket->isconnected )
   {
      pthistask->mylasterror = WSAEISCONN ;
      return SOCKET_ERROR ;
   }
   else
   {
      pthissocket->isconnected = TRUE ;
   }

   pthissocket->socketfamily = name->sa_family ;

// permit the user to have a WINMOCK.INI file which can have the following format
// [connect]
// rc=<error code>
   rc = GetPrivateProfileString( "lasterror", "connect",
          "",
          szErrorString, ERRORSTRINGSIZE,
          "WINMOCK.INI" ) ;
   if ( 0 == rc )
   {
      rc = GetPrivateProfileString( "connect", "rc",
             "ZERO",
             szErrorString, ERRORSTRINGSIZE,
             "WINMOCK.INI" ) ;

      rc = wmConvertErrorString( szErrorString ) ;
   }
   else
   {
      rc = wmConvertErrorString( szErrorString ) ;
      pthistask->mylasterror = rc ;
      return( SOCKET_ERROR ) ;
   }

   return( rc ) ;
}

// get host information for the given host name
//   Note -- one copy of HOSTENT structure is allocated per task
LPHOSTENT APIENTRY gethostbyname( LPSTR hostname )
{
   PMYTASK   pthistask ;
   LPHOSTENT lpHostEnt ;
   HOSTENT *pHE ;
   HLOCAL hLMem ;
   HTASK  ht = GetCurrentTask() ;
   int rc ;
static LPSTR lpszHostAddr ; // put this in DS so it stays valid for caller
                            //   (as they may access it only long after return)
                            //   (a pointer to a value on the stack
                            //   would have changed by then)

   DebugBreak() ;

   // if this app isn't on our task list, then they
   //    haven't called WSAStartup() or have already called WSACleanup()
   pthistask = wmGetTaskInfo( ht ) ;
   if ( NULL == pthistask )
   {
     return( NULL ) ;
   }

   if ( NULL == pthistask->myhostent )
   {
      // HACK!  Get enough space to store an address together with
      //   the hostent allocation
      hLMem = LocalAlloc( sizeof( HOSTENT ) + ADDRBUFFERSIZE, LPTR ) ;
      pHE = (HOSTENT *) LocalLock ( hLMem ) ;
      lpHostEnt = (LPHOSTENT) pHE ;
      pthistask->myhostent =lpHostEnt ;
      lpszHostAddr = (LPSTR) lpHostEnt + sizeof( HOSTENT ) ;
   }
   else
   {
      lpHostEnt = pthistask->myhostent ;
      lpszHostAddr = (LPSTR) lpHostEnt + sizeof( HOSTENT ) ;
   }

// permit the user to have a WINMOCK.INI file which can have the following format
// NOTE:  WINSOCK constants must be resolved in this file -- use 2, not PF_INET
// [gethostbyname]
// h_name -- ignored, filled in from call
// h_addrtype=2 (PF_INET)
// h_length=4
// h_addr_list -- RESTRICTION -- only a single addr list entry is permitted
   lpHostEnt->h_name = pthistask->szNameSpace ;
   wmfstrcpy( lpHostEnt->h_name, hostname ) ;

   lpHostEnt->h_addrtype = GetPrivateProfileInt(
          "gethostbyname", "h_addrtype", PF_INET, "WINMOCK.INI" ) ;
   lpHostEnt->h_length = GetPrivateProfileInt(
          "gethostbyname", "h_length", 4, "WINMOCK.INI" ) ;
   GetPrivateProfileString( "gethostbyname", "h_addrlist",
          "11.22.33.44",
          lpszHostAddr, ADDRBUFFERSIZE,
          "WINMOCK.INI" ) ;
   lpHostEnt->h_addr_list = (char FAR * FAR * ) &lpszHostAddr ;


   rc = GetPrivateProfileString( "lasterror", "gethostbyname",
          "",
          szErrorString, ERRORSTRINGSIZE,
          "WINMOCK.INI" ) ;
   if ( 0 == rc )
   {
   // currently we do not give the user the ability to modify the RC for this call
      return( lpHostEnt ) ;
   }
   else
   {
      rc = wmConvertErrorString( szErrorString ) ;
      pthistask->mylasterror = rc ;
      return( NULL ) ;
   }

}

// gets standard host name for local address
int APIENTRY gethostname ( LPSTR hostname, int namelen )
{
   HTASK ht = GetCurrentTask() ;
   PMYTASK   pthistask ;
   int rc ;

   DebugBreak() ;

   // if this app isn't on our task list, then they
   //    haven't called WSAStartup() or have already called WSACleanup()
   pthistask = wmGetTaskInfo( ht ) ;
   if ( NULL == pthistask )
   {
      return( SOCKET_ERROR ) ;
   }

// permit the user to have a WINMOCK.INI file which can have the following format
// [gethostname]
// hostname
// rc=<error code>
   GetPrivateProfileString( "gethostname", "hostname",
          "myhostname",
          hostname, namelen,
          "WINMOCK.INI" ) ;

   rc = GetPrivateProfileString( "lasterror", "gethostname",
          "",
          szErrorString, ERRORSTRINGSIZE,
          "WINMOCK.INI" ) ;
   if ( 0 == rc )
   {
      rc = GetPrivateProfileString( "gethostname", "rc",
             "ZERO",
             szErrorString, ERRORSTRINGSIZE,
             "WINMOCK.INI" ) ;

      rc = wmConvertErrorString( szErrorString ) ;
   }
   else
   {
      rc = wmConvertErrorString( szErrorString ) ;
      pthistask->mylasterror = rc ;
      return( SOCKET_ERROR ) ;
   }

   return( rc ) ;
}

// get protocol information for the given protocol name
//   Note:  currently only recognizes TCP
LPPROTOENT APIENTRY getprotobyname( LPSTR protocolname )
{
   HTASK ht = GetCurrentTask() ;
   PMYTASK pthistask ;
   LPPROTOENT lpProtoEnt ;
   HLOCAL hLMem ;
   PROTOENT *pPE ;
   int rc ;

   DebugBreak() ;

   // if this app isn't on our task list, then they
   //    haven't called WSAStartup() or have already called WSACleanup()
   pthistask = wmGetTaskInfo( ht ) ;
   if ( NULL == pthistask )
   {
     return( NULL ) ;
   }

   if ( NULL == pthistask->myprotoent )
   {
      hLMem = LocalAlloc( sizeof( PROTOENT ), LPTR ) ;
      pPE = (PROTOENT *) LocalLock ( hLMem ) ;
      lpProtoEnt = (LPPROTOENT) pPE ;
      pthistask->myprotoent = lpProtoEnt ;
   }
   else
   {
      lpProtoEnt = pthistask->myprotoent ;
   }

// permit the user to have a WINMOCK.INI file which can have the following format
// [getprotobyname]
// p_proto
   lpProtoEnt->p_name = pthistask->szNameSpace ;
   wmfstrcpy( lpProtoEnt->p_name, protocolname ) ;
   lpProtoEnt->p_aliases = NULL ;
   lpProtoEnt->p_proto = GetPrivateProfileInt(
          "getprotobyname", "p_proto", IPPROTO_TCP, "WINMOCK.INI" ) ;

//   pthistask->mylasterror = WSANO_DATA ; // valid name, no data record

   rc = GetPrivateProfileString( "lasterror", "getprotobyname",
          "",
          szErrorString, ERRORSTRINGSIZE,
          "WINMOCK.INI" ) ;
   if ( 0 == rc )
   {
   // currently we do not give the user the ability to modify the RC for this call
      return( lpProtoEnt ) ;
   }
   else
   {
      rc = wmConvertErrorString( szErrorString ) ;
      pthistask->mylasterror = rc ;
      return( NULL ) ;
   }
}

// for now we assume everything is http... a real implementation would
//   check the services file in your ETC directory...
// one copy of SERVENT structure is allocated per thread
// NOTE:  need to clean this up
LPSERVENT APIENTRY getservbyname( LPSTR servicename, LPSTR proto )
{
   HTASK ht = GetCurrentTask() ;
   PMYTASK pthistask ;
   LPSERVENT lpServEnt ;
   HLOCAL hLMem ;
   SERVENT * pSE ;
   int rc ;

   DebugBreak() ;

   // if this app isn't on our task list, then they
   //    haven't called WSAStartup() or have already called WSACleanup()
   pthistask = wmGetTaskInfo( ht ) ;
   if ( NULL == pthistask )
   {
     return( NULL ) ;
   }

   if ( NULL == pthistask->myservent )
   {
      hLMem = LocalAlloc( sizeof( SERVENT ), LPTR ) ;
      pSE = (SERVENT *) LocalLock ( hLMem ) ;
      lpServEnt = (LPSERVENT) pSE ;
      pthistask->myservent = lpServEnt ;
   }
   else
   {
      lpServEnt = pthistask->myservent ;
   }

// expand this to use "nntp" and "ftp" at a minimum
   lpServEnt->s_name = "Hypertext Transaction Protocol" ;
   // note that the following is in network byte order
   lpServEnt->s_port = 0x8000;
   lpServEnt->s_proto = "http" ;

// permit the user to have a WINMOCK.INI file which can have the following format
// [getservbyname]
// s_proto
// s_port
   lpServEnt->s_name = servicename ; // we reuse the str passed in by the caller
   GetPrivateProfileString( "getservbyname", "s_proto",
          "",
          pthistask->szNameSpace, SIZEOFNAMESPACE,
          "WINMOCK.INI" ) ;
   lpServEnt->s_proto = pthistask->szNameSpace ;
   lpServEnt->s_aliases = NULL ;
   lpServEnt->s_port = GetPrivateProfileInt(
          "getservbyname", "s_port", IPPROTO_TCP, "WINMOCK.INI" ) ;

   rc = GetPrivateProfileString( "lasterror", "getservbyname",
          "",
          szErrorString, ERRORSTRINGSIZE,
          "WINMOCK.INI" ) ;
   if ( 0 == rc )
   {
   // currently we do not give the user the ability to modify the RC for this call
      return( lpServEnt ) ;
   }
   else
   {
      rc = wmConvertErrorString( szErrorString ) ;
      pthistask->mylasterror = rc ;
      return( NULL ) ;
   }
}


// this function will read a file name out of the winmock.ini file
// the contents of the file will then be read into the buffer passed in
// reads data off of the connected datagram or stream socket
int APIENTRY recv ( SOCKET s, LPSTR buf, int len, int flags )
{
   HTASK ht = GetCurrentTask() ;
   PMYTASK pthistask ;
   PMYSOCKET pthissocket = wmGetSocketInfo( s ) ;
   int rc, errorrc ;
   WORD wLeftToRecv ;
   LPSTR lpszRecvText ;
   HANDLE hRecv ;

   DebugBreak() ;

   // if this app isn't on our task list, then they
   //    haven't called WSAStartup() or have already called WSACleanup()
   pthistask = wmGetTaskInfo( ht ) ;
   if ( NULL == pthistask )
   {
     return( NULL ) ;
   }

   if ( NULL == pthissocket )
   {
      pthistask->mylasterror = WSAENOTSOCK ;
      return( SOCKET_ERROR ) ;
   }

   // if socket doesn't belong to this task, then return error
   if ( ht != pthissocket->mytask )
   {
      pthistask->mylasterror = WSAENOTSOCK ;
      return SOCKET_ERROR ;
   }

   // if socket is not connected, then error
   //   in a implementation with a network back end, you would also need to check
   //   if the remote connection has reset the virtual circuit
   //   and return WSAECONNRESET, if so
   if ( FALSE == pthissocket->isconnected )
   {
      pthistask->mylasterror = WSAENOTCONN ;
      return SOCKET_ERROR ;
   }

   if ( SOCK_DGRAM == pthissocket->sockettype )
   {
      // verify that caller buffer is big enough for packet size or
      //   return WSAEMSGSIZE and as much as will fit
      //   Note:  not done in this version
   }

// TRUE == isstillreceiving means that there is data available on this socket
   if ( TRUE == pthissocket->isstillreceiving )
   {
// permit the user to have a WINMOCK.INI file which can have the following format
// [recv]
// buffer
      if ( NULL == pthissocket->hRecv )
      {
         rc = GetPrivateProfileString( "recv", "buffer",
             "* no received text *",
             buf, len,
             "WINMOCK.INI" ) ;

         // did the user give us a file name?
         if ( '.' == buf[ 0 ] )
         {
             // the . in first position marks this as a file name to
             //   be read into the user buffer
             hRecv = wmLoadFile( &buf[ 1 ] ) ; // skip leading .

             if ( NULL == hRecv ) // file could not be loaded
             {
                pthistask->mylasterror = WSAECONNABORTED ;
                return SOCKET_ERROR ;
             }

         }
         else // they gave us a string...
         {
             LPSTR lpszBuffer ;

             hRecv = GlobalAlloc( GMEM_MOVEABLE, rc ) ;
             lpszBuffer = GlobalLock( hRecv ) ;
             wmfmemcpy( lpszBuffer, buf, rc ) ;
         }

         pthissocket->hRecv = hRecv ;
         pthissocket->wRecvSize = (WORD) GlobalSize( hRecv ) ;
         pthissocket->wRecvOffset = 0 ;
      }

      wLeftToRecv = pthissocket->wRecvSize - pthissocket->wRecvOffset ;
      if ( wLeftToRecv < (WORD) len)
      {
         // we can read everything in one chunk so stop receiving
         rc = wLeftToRecv ;
         pthissocket->isstillreceiving = FALSE ;
      }
      else  // otherwise return as much to the user as will fit in their buffer
      {
         rc = len ;
      }

      lpszRecvText = GlobalLock( pthissocket->hRecv ) ;
      wmfmemcpy( buf, &lpszRecvText[ pthissocket->wRecvOffset ], rc ) ;
      pthissocket->wRecvOffset += len ;
      GlobalUnlock( pthissocket->hRecv ) ;
   }
   else
   { // else no data:  we assume that the connection has been terminated
     //   a network implementation may need to block here until
     //   data can be received from the remote end...
      rc = 0 ; // stop receive loop
      pthissocket->wRecvOffset = 0 ;           // start at beginning of
                                             //   buffer next time
      pthissocket->isstillreceiving = TRUE ; // prepare to receive again
   }

   errorrc = GetPrivateProfileString( "lasterror", "recv",
          "",
          szErrorString, ERRORSTRINGSIZE,
          "WINMOCK.INI" ) ;
   if ( 0 == errorrc )
   {
   // currently we do not give the user the ability to modify the RC for this call
      return( rc ) ;
   }
   else
   {
      rc = wmConvertErrorString( szErrorString ) ;
      pthistask->mylasterror = rc ;
      return( SOCKET_ERROR ) ;
   }
}

// reads data off of the (possibly) connected socket and
//   return the remote address
// from and fromlen parms are ignored on stream sockets
int APIENTRY recvfrom ( SOCKET s, LPSTR buf, int len, int flags,
                         LPSOCKADDR from, LPINT fromlen )
{
   HTASK ht = GetCurrentTask() ;
   PMYTASK pthistask ;
   PMYSOCKET pthissocket = wmGetSocketInfo( s ) ;
   int rc, errorrc ;
   LPSTR ptext = "This is received text.\r\n.\r\n" ;
   WORD wLeftToRecv ;
   LPSTR lpszRecvText ;
   HANDLE hRecv ;

   DebugBreak() ;

   // if this app isn't on our task list, then they
   //    haven't called WSAStartup() or have already called WSACleanup()
   pthistask = wmGetTaskInfo( ht ) ;
   if ( NULL == pthistask )
   {
     return( NULL ) ;
   }

   if ( NULL == pthissocket )
   {
      pthistask->mylasterror = WSAENOTSOCK ;
      return( SOCKET_ERROR ) ;
   }

   // if socket doesn't belong to this task, then return error
   if ( ht != pthissocket->mytask )
   {
      pthistask->mylasterror = WSAENOTSOCK ;
      return SOCKET_ERROR ;
   }

// NOTE:  only STREAM sockets need to be connected and only
//   DGRAM sockets have an use the from field so...

// if the caller has passed a non-NULL from param, then they want
//   specific information on the other side of the conversation
   if ( SOCK_DGRAM == pthissocket->sockettype )
   {
      if ( from )
      {
         LPSOCKADDR_IN lpsa_in = (LPSOCKADDR_IN) from ;

         // if the user from structure isn't begin enough, then return
         //   an error.  Otherwise, copy how big our structure is so
         //   the caller will know
         if ( *fromlen < sizeof(SOCKADDR_IN) )
         {
            pthistask->mylasterror = WSAEFAULT ;
            return( SOCKET_ERROR ) ;
         }
         else
         {
            *fromlen = sizeof( SOCKADDR_IN ) ;
         }

         lpsa_in->sin_family = GetPrivateProfileInt(
             "recvfrom", "sin_family", AF_INET, "WINMOCK.INI" ) ;
         lpsa_in->sin_port = GetPrivateProfileInt(
             "recvfrom", "sin_port", 0, "WINMOCK.INI" ) ;
         lpsa_in->sin_addr.s_addr = 0x11223344 ;
      }
   }
   else
   {  // this is a STREAM socket
      // if socket is not connected, then error
      //   in a real implementation, you would also need to check
      //   if the remote connection has reset the virtual circuit
      //   and return WSAECONNRESET (stream sockets only)
      if ( FALSE == pthissocket->isconnected )
      {
         pthistask->mylasterror = WSAENOTCONN ;
         return SOCKET_ERROR ;
      }
   }

// TRUE == isstillreceiving means that there is data available on this socket
   if ( TRUE == pthissocket->isstillreceiving )
   {
// permit the user to have a WINMOCK.INI file which can have the following format
// [recvfrom]
// sin_family
// sin_port
// buffer
      if ( NULL == pthissocket->hRecv )
      {
         rc = GetPrivateProfileString( "recv", "buffer",
             "* no received text *",
             buf, len,
             "WINMOCK.INI" ) ;

         // did the user give us a file name?
         if ( '.' == buf[ 0 ] )
         {
             // the . in first position marks this as a file name to
             //   be read into the user buffer
             hRecv = wmLoadFile( &buf[ 1 ] ) ; // skip leading .

             if ( NULL == hRecv ) // file could not be loaded
             {
                pthistask->mylasterror = WSAECONNABORTED ;
                return SOCKET_ERROR ;
             }

         }
         else // they gave us a string...
         {
             LPSTR lpszBuffer ;

             hRecv = GlobalAlloc( GMEM_MOVEABLE, rc ) ;
             lpszBuffer = GlobalLock( hRecv ) ;
             wmfmemcpy( lpszBuffer, buf, rc ) ;
         }

         pthissocket->hRecv = hRecv ;
         pthissocket->wRecvSize = (WORD) GlobalSize( hRecv ) ;
         pthissocket->wRecvOffset = 0 ;
      }

      wLeftToRecv = pthissocket->wRecvSize - pthissocket->wRecvOffset ;
      if ( wLeftToRecv < (WORD) len)
      {
         // we can read everything in one chunk so stop receiving
         rc = wLeftToRecv ;
         pthissocket->isstillreceiving = FALSE ;
      }
      else  // otherwise return as much to the user as will fit in their buffer
      {
         rc = len ;
      }

      lpszRecvText = GlobalLock( pthissocket->hRecv ) ;
      wmfmemcpy( buf, &lpszRecvText[ pthissocket->wRecvOffset ], rc ) ;
      pthissocket->wRecvOffset += len ;
      GlobalUnlock( pthissocket->hRecv ) ;
   }
   else
   { // else no data:  we assume that the connection has been terminated
     //   a network implementation may need to block here until
     //   data can be received from the remote end...
      rc = 0 ; // stop receive loop
      pthissocket->wRecvOffset = 0 ;           // start at beginning of
                                             //   buffer next time
      pthissocket->isstillreceiving = TRUE ; // prepare to receive again
   }


   errorrc = GetPrivateProfileString( "lasterror", "recvfrom",
          "",
          szErrorString, ERRORSTRINGSIZE,
          "WINMOCK.INI" ) ;
   if ( 0 == errorrc )
   {
   // currently we do not give the user the ability to modify the RC for this call
      return( rc ) ;
   }
   else
   {
      rc = wmConvertErrorString( szErrorString ) ;
      pthistask->mylasterror = rc ;
      return( SOCKET_ERROR ) ;
   }
}

// Note that this is a NOP in our version
// Sends data on a connected socket.  For DGRAM sockets, you
//   cannot exceed the maximum packet size
int APIENTRY send ( SOCKET s, LPSTR buf, int len, int flags )
{
   HTASK ht = GetCurrentTask() ;
   PMYTASK pthistask ;
   PMYSOCKET pthissocket = wmGetSocketInfo( s ) ;
   int rc, errorrc ;

   DebugBreak() ;

   // if this app isn't on our task list, then they
   //    haven't called WSAStartup() or have already called WSACleanup()
   pthistask = wmGetTaskInfo( ht ) ;
   if ( NULL == pthistask )
   {
     return( NULL ) ;
   }

   if ( NULL == pthissocket )
   {
      pthistask->mylasterror = WSAENOTSOCK ;
      return( SOCKET_ERROR ) ;
   }

   // if socket doesn't belong to this task, then return error
   if ( ht != pthissocket->mytask )
   {
      pthistask->mylasterror = WSAENOTSOCK ;
      return SOCKET_ERROR ;
   }

   // if socket is not connected, then error
   //   in a real implementation, you would also need to check
   //   if the remote connection has reset the virtual circuit
   //   and return WSAECONNRESET (stream sockets only)
   if ( FALSE == pthissocket->isconnected )
   {
      pthistask->mylasterror = WSAENOTCONN ;
      return SOCKET_ERROR ;
   }

   if ( SOCK_DGRAM == pthissocket->sockettype )
   {
      if ( len > MAXUDPDG ) // maximum IP packet length exceeded so don't send
      {
        pthistask->mylasterror = WSAEMSGSIZE ;
        return SOCKET_ERROR ;
      }
   }

   // in a network-aware sockets version of WinMock, we'd need
   //   to actually send the data, now we just return the number
   //   of characters sent and mark the socket as already having
   //   received a reply... (isstillreceiving=TRUE)
   rc = len ;
   pthissocket->isstillreceiving = TRUE ;

   errorrc = GetPrivateProfileString( "lasterror", "send",
          "",
          szErrorString, ERRORSTRINGSIZE,
          "WINMOCK.INI" ) ;
   if ( 0 == errorrc )
   {
   // currently we do not give the user the ability to modify the RC for this call
      return( rc ) ;
   }
   else
   {
      rc = wmConvertErrorString( szErrorString ) ;
      pthistask->mylasterror = rc ;
      return( SOCKET_ERROR ) ;
   }
}

// writes data out on a socket (does not imply that the data were delivered)
//   for stream sockets, this is equivalent to send(), and the to address
//   is ignored.
int APIENTRY sendto ( SOCKET s, LPSTR buf, int len, int flags,
                       LPSOCKADDR to, int tolen )
{
   HTASK ht = GetCurrentTask() ;
   PMYTASK pthistask ;
   PMYSOCKET pthissocket = wmGetSocketInfo( s ) ;
   int rc, errorrc ;

   DebugBreak() ;

   // if this app isn't on our task list, then they
   //    haven't called WSAStartup() or have already called WSACleanup()
   pthistask = wmGetTaskInfo( ht ) ;
   if ( NULL == pthistask )
   {
     return( NULL ) ;
   }

   if ( NULL == pthissocket )
   {
      pthistask->mylasterror = WSAENOTSOCK ;
      return( SOCKET_ERROR ) ;
   }

   // if socket doesn't belong to this task, then return error
   if ( ht != pthissocket->mytask )
   {
      pthistask->mylasterror = WSAENOTSOCK ;
      return SOCKET_ERROR ;
   }

   // if STREAM socket is not connected, then error
   //   in a real implementation, you would also need to check
   //   if the remote connection has reset the virtual circuit
   //   and return WSAECONNRESET (stream sockets only)
   if ( SOCK_STREAM == pthissocket->sockettype )
   {
      if ( FALSE == pthissocket->isconnected )
      {
         pthistask->mylasterror = WSAENOTCONN ;
         return SOCKET_ERROR ;
      }
   }
   else
   { // this is SOCK_DGRAM so process the TO field
     //   not needed here as we're not actually sending anything
      if ( len > MAXUDPDG ) // maximum IP packet length exceeded so don't send
      {
        pthistask->mylasterror = WSAEMSGSIZE ;
        return SOCKET_ERROR ;
      }
   }

   // in a network-aware sockets version of WinMock, we'd need
   //   to actually send the data, now we just return the number
   //   of characters sent and mark the socket as already having
   //   received a reply... (isstillreceiving=TRUE)
   rc = len ;
   pthissocket->isstillreceiving = TRUE ;

   errorrc = GetPrivateProfileString( "lasterror", "sendto",
          "",
          szErrorString, ERRORSTRINGSIZE,
          "WINMOCK.INI" ) ;
   if ( 0 == errorrc )
   {
   // currently we do not give the user the ability to modify the RC for this call
      return( rc ) ;
   }
   else
   {
      rc = wmConvertErrorString( szErrorString ) ;
      pthistask->mylasterror = rc ;
      return( SOCKET_ERROR ) ;
   }
}

// allocates a socket, an endpoint of communication
SOCKET APIENTRY socket ( int af, int type, int protocol )
{
   SOCKET s ;
   HTASK ht ;
   PMYTASK pthistask ;
   PMYSOCKET pthissocket ;
   int rc ;

   DebugBreak() ;

   ht = GetCurrentTask() ;

   // get task info for current task
   pthistask = wmGetTaskInfo( ht ) ;
   if ( NULL == pthistask )
   {
     return( NULL ) ;
   }

   // only ARPA internet address format is supported
   if ( af && ( AF_INET != af ) )
   {
      pthistask->mylasterror = WSAEAFNOSUPPORT ;
      return INVALID_SOCKET ;
   }
   else
   { // replace AF_UNSPEC with AF_INET
      af = AF_INET ;
   }

   // get next free socket
   pthissocket = wmGetFreeSocket() ;

   // if all sockets are taken up, then fail this operation
   if ( NULL == pthissocket )
   {
      pthistask->mylasterror = WSAEMFILE ;
      return ( (SOCKET) INVALID_SOCKET ) ; // return error
   }

   switch( protocol )
   {
      case IPPROTO_TCP:
        switch( type )
        {
           case SOCK_STREAM:
             break ;
           case 0:
             type = SOCK_STREAM ;
             break ;
           case SOCK_DGRAM:  // fall through
           default:
             pthistask->mylasterror = WSAEPROTOTYPE ;
             wmClearSocketStruct( pthissocket ) ; // free socket we just alloced
             return INVALID_SOCKET ;
        }
        break ;
      case IPPROTO_UDP:
        switch( type )
        {
           case SOCK_DGRAM:
             break ;
           case 0:
             type = SOCK_DGRAM ;
             break ;
           case SOCK_STREAM:      // fall through
           default:
             pthistask->mylasterror = WSAEPROTOTYPE ;
             wmClearSocketStruct( pthissocket ) ; // free socket we just alloced
             return INVALID_SOCKET ;
        }
        break ;
      case 0:
        switch( type )
        {
           case SOCK_DGRAM:
             protocol = IPPROTO_UDP ;
             break ;
           case SOCK_STREAM:
             protocol = IPPROTO_TCP ;
             break ;
           case 0:
           default:
             pthistask->mylasterror = WSAEPROTOTYPE ;
             wmClearSocketStruct( pthissocket ) ; // free socket we just alloced
             return INVALID_SOCKET ;
        }
        break ;
        default:
          pthistask->mylasterror = WSAEPROTONOSUPPORT ;
          wmClearSocketStruct( pthissocket ) ; // free socket we just alloced
          return INVALID_SOCKET ;
   }

   // set socket flags
   pthissocket->socketprotocol = protocol ;
   pthissocket->sockettype = type ;
   pthissocket->socketfamily = af ;
   pthissocket->mytask = pthistask->mytask ;

   s = (SOCKET) pthissocket->mys ; // index of our socket element

   rc = GetPrivateProfileString( "lasterror", "socket",
          "",
          szErrorString, ERRORSTRINGSIZE,
          "WINMOCK.INI" ) ;
   if ( 0 == rc )
   {
   // we don't let the user modify the normal return code
      return( s ) ;
   }
   else
   {
      rc = wmConvertErrorString( szErrorString ) ;
      pthistask->mylasterror = rc ;
      return( INVALID_SOCKET ) ;
   }

}

// gets error code for last windows sockets operation that failed
//   for this task
int APIENTRY WSAGetLastError( void )
{
   int rc ;
   HTASK ht = GetCurrentTask() ;
   PMYTASK pthistask ;

   // get the task-specific error
   if ( pthistask = wmGetTaskInfo( ht ) )
   {
      return( pthistask->mylasterror ) ;
   }
   else // this task has not called WSAStartup() or has called WSACleanup()
   {
      return( WSANOTINITIALISED ) ;
   }

   return( rc ) ;
}

// deregister from WinSock (we use this to clean up allocated resources)
//   Note that this api must be called once for every WSAStartup() call
int APIENTRY WSACleanup( void )
{
   int rc ;
   HTASK ht = GetCurrentTask() ;
   PMYTASK pthistask = wmGetTaskInfo( ht ) ;

   DebugBreak() ;

   // if not known task then WSAStartup() not called
   //   or WSACleanup() already called() -- fail!
   if ( NULL == pthistask )
   {  // GetLastError() will return WSANOTINITIALISED, if called
      return( SOCKET_ERROR ) ;
   }

   // if task is blocking then fail WSACleanup()
   if ( TRUE == pthistask->isblocking )
   {
      pthistask->mylasterror = WSAEINPROGRESS ;
      return( SOCKET_ERROR ) ;
   }

   ntasks -= 1 ; // decrement local task count

   // now remove this task from the task list
   // so that future WinSock calls will fail
   // (if this is the last call to WSACleanup() )
   wmRemoveTaskStruct( pthistask ) ;

   rc = 0 ; // success
   return( rc ) ;
}

// must be first WinSock call made by an application
// may be called multiple times by (or for) the same task
// The WinSock specification permits last error codes to be
//   returned by this routine as per-task structures for last
//   error handling aren't yet set up for this task if the call fails
int APIENTRY WSAStartup( WORD wVersionRequested, LPWSADATA lpWSAData )
{
   int rc = 0 ;                     // success
   HTASK ht = GetCurrentTask() ;
   int wVersion = GetPrivateProfileInt(
          "wsastartup", "wVersion", MAKEWORD( 1, 1 ), "WINMOCK.INI" ) ;

   DebugBreak() ;

   // error if they want a higher version of the windows specification
   if ( HIBYTE( wVersionRequested ) > HIBYTE( wVersion ) )
   {
      return( WSAVERNOTSUPPORTED ) ;
   }

   // error if they have the same major version number but a higher minor
   //   version number
   if ( ( HIBYTE( wVersionRequested ) == HIBYTE( wVersion ) ) &&
        ( LOBYTE( wVersionRequested ) > LOBYTE( wVersion ) )
      )
   {
      return( WSAVERNOTSUPPORTED ) ;
   }


   ntasks += 1 ; // local count of number of tasks

   // add this task as a new WinSock task
   wmAddTask( ht ) ;

// permit the user to have a WINMOCK.INI file which can have the following format
// [wsastartup]
// wVersion=101
// wHighVersion=101
// szDescription=...
// szSystemStatus=...
   lpWSAData->wVersion = wVersion ;
   lpWSAData->wHighVersion = GetPrivateProfileInt(
          "wsastartup", "wHighVersion", wVersion, "WINMOCK.INI" ) ;
   GetPrivateProfileString( "wsastartup", "szDescription",
          "WinMock WinSock Off-Line Tester",
          lpWSAData->szDescription, sizeof( lpWSAData->szDescription ),
          "WINMOCK.INI" ) ;
   GetPrivateProfileString( "wsastartup", "szSystemStatus",
          "",
          lpWSAData->szSystemStatus, sizeof( lpWSAData->szSystemStatus ),
          "WINMOCK.INI" ) ;

   // don't permit user to override MAXSOCKETS in this version
   lpWSAData->iMaxSockets = MAXSOCKETS ;
   lpWSAData->iMaxUdpDg = MAXUDPDG ; // 512 -- windows sockets smallest permitted size
                                //   for this field
   lpWSAData->lpVendorInfo = NULL ;

   return( rc ) ;
}

