//    WINMOCK -- Windows Sockets Non-Network Version for testing
//      sockets apps off-line
// WinUtil.C -- various utility functions
// Copyright (C) Joseph Hlavaty, 1995.  All rights reserved.

#include <windows.h>
#include <winsock.h>
#include "winmock.h"

// the following are far pointer versions of C library functions
// as we're a small model DLL
LPSTR wmfstrcpy( LPSTR pd, LPSTR ps )
{
   LPSTR pdeststart = pd ;

   while ( *ps )
   {
      *pd++ = *ps++ ;
   }

   return( pdeststart ) ;
}

int   wmfstrlen( LPSTR ps )
{
   int count = 0 ;

   while( *ps )
   {
      ps++ ;
      count++ ;
   }

   return( count ) ;
}

LPSTR wmfmemcpy( LPSTR pd, LPSTR ps, unsigned int count )
{
   LPSTR pdorig = pd ;

   while( count-- )
   {
      *pd++ = *ps ;

      if ('\0' == *ps)
      {
         *pd = '\0' ;
         return( pdorig ) ;
      }
      else
      {
         ps++ ;
      }
   }
   *pd='\0' ;

   return( pdorig ) ;
}

int wmfstrcmp( LPSTR s1, LPSTR s2 )
{
   int rc = 0 ; // identical for lengths compared

   for ( ; *s1 == *s2 ; s1++, s2++ )
   {
      if ( '\0' == *s1 )
      {
         return( rc ) ;
      }

   }

   rc = *s1 - *s2 ;

   return( rc ) ;

}

WORD  wmConvertErrorString( LPSTR lpszErrorString )
{
   int i ;

   for (i=0; i < ERRORLISTSIZE ; i++)
   {  // is our error string the same as the string in this entry
      //    of the error table (note that case is preserved during the test)
      if ( !wmfstrcmp( lpszErrorString, ErrorList[i].pszError ) )
      {
         // yes, we found our error msg
         return ErrorList[i].iError ;

      }
   }
}

WORD wmInitializeTaskList()
{
   PMYTASK pthistask = pataskshead ;
   int i ;

   // mark all our tasks as unused
   for ( i = 0; i < MAXTASKS; i++)
   {
      pthistask[i].mytask = (HTASK) NULL ;
      pthistask[i].myhostent = NULL ;
      pthistask[i].myprotoent = NULL ;
      pthistask[i].myservent = NULL ;
      pthistask[i].isblocking = FALSE ;
      pthistask[i].mylasterror = 0 ;
      pthistask[i].WSAStartupCount = 0 ;
      pthistask[i].ptnext = NULL ;
   }

   return( TRUE ) ;
}

PMYTASK wmGetTaskInfo( HTASK ht )
{
   PMYTASK   pthistask = pataskshead ;
   int iTaskOffset ;

   for (iTaskOffset = 0; iTaskOffset < MAXTASKS ; iTaskOffset++)
   {
       if ( ht == pthistask[iTaskOffset].mytask )
          return( &pthistask[iTaskOffset] ) ;
   }

   return( NULL ) ;
}

PMYTASK wmAddTask( HTASK ht )
{
   PMYTASK pthistask = pataskshead ;
   int iTaskOffset ;

   for (iTaskOffset = 0; iTaskOffset < MAXTASKS ; iTaskOffset++)
   {
       if ( NULL == pthistask[iTaskOffset].mytask )
       {
          pthistask[iTaskOffset].mytask = ht ;
          pthistask[iTaskOffset].myhostent = NULL ;
          pthistask[iTaskOffset].myprotoent = NULL ;
          pthistask[iTaskOffset].myservent = NULL ;
          pthistask[iTaskOffset].isblocking = FALSE ;
          pthistask[iTaskOffset].WSAStartupCount++ ;
          pthistask[iTaskOffset].mylasterror = 0 ;
          return( &pthistask[iTaskOffset] ) ;
       }
   }

   return( NULL ) ;
}

// the caller is assumed to be trusted code (it is called with the
//   WinMock library) calling with a valid pointer
void wmRemoveTaskStruct( PMYTASK pthistask )
{
   HTASK ht ;
   HLOCAL hLMem ;
   int WSACount ;

   if ( pthistask)
   {

      WSACount = --(pthistask->WSAStartupCount) ;

      // an application must call WSACleanup() as many times as it (or someone
      //   else) has called WSAStartup() so don't free resources until the
      //   count of calls falls to zero
      if ( WSACount )
      {
         return ;
      }

      // fall through if count is zero and dealloc resources
      ht = pthistask->mytask ; // remember task id

      pthistask->mytask = NULL ;
      pthistask->isblocking = FALSE ;
      pthistask->mylasterror = 0 ;

      if ( pthistask->myhostent )
      {
         hLMem = LocalHandle ( (WORD) pthistask->myhostent ) ;
         LocalUnlock ( hLMem ) ;
         LocalFree( hLMem ) ;
         pthistask->myhostent = NULL ;
      }

      if ( pthistask->myprotoent )
      {
         hLMem = LocalHandle ( (WORD) pthistask->myprotoent ) ;
         LocalUnlock ( hLMem ) ;
         LocalFree( hLMem ) ;
         pthistask->myprotoent = NULL ;
      }

      if ( pthistask->myservent )
      {
         hLMem = LocalHandle ( (WORD) pthistask->myservent ) ;
         LocalUnlock ( hLMem ) ;
         LocalFree( hLMem ) ;
         pthistask->myservent = NULL ;
      }

      // iterate through socket array, freeing all sockets owned by this app
      //   for use by another task
      wmClearSocketsForThisTask( ht ) ;
   }
}

WORD wmInitializeSocketList()
{
   PMYSOCKET pthissocket = pasocketshead ;
   int i ;

   // mark all our sockets as unused...
   for ( i = 0; i < MAXSOCKETS; i++)
   {
      pthissocket[i].mys = SOCKET_ERROR ;
      pthissocket[i].mytask = (HTASK) NULL ;
      pthissocket[i].isconnected = FALSE ;
      pthissocket[i].isstillreceiving = FALSE ;
      pthissocket[i].socketblocking = FALSE ;
      pthissocket[i].socketfamily = 0 ;
      pthissocket[i].socketprotocol = 0 ;
      pthissocket[i].sockettype = 0 ;
      pthissocket[i].hRecv = NULL ;
      pthissocket[i].wRecvSize = 0 ;
      pthissocket[i].wRecvOffset = 0 ;
      pthissocket[i].hSend = NULL ;
      pthissocket[i].psnext = NULL ;
   }

  return( TRUE ) ;
}

PMYSOCKET wmGetSocketInfo( SOCKET s )
{
   PMYSOCKET pthissocket = pasocketshead ;
   int iSocketOffset ;

   for (iSocketOffset = 0; iSocketOffset < MAXTASKS ; iSocketOffset++)
   {
       if ( s == pthissocket[iSocketOffset].mys )
          return( &pthissocket[iSocketOffset] ) ;
   }

   return( NULL ) ;
}

PMYSOCKET wmGetFreeSocket( )
{
   PMYSOCKET pthissocket = pasocketshead ;
   int iSocketOffset ;

   for (iSocketOffset = 0; iSocketOffset < MAXSOCKETS ; iSocketOffset++)
   {
       // if this socket is not assigned to a task, then it is available
       if ( NULL == pthissocket[iSocketOffset].mytask )
       {
          // set socket value for caller to extract
          pthissocket[iSocketOffset].mys = (SOCKET) iSocketOffset ;
          return( &pthissocket[iSocketOffset] ) ;
       }
   }

   // all sockets are taken, return error
   return( NULL ) ;
}

// note that in an implementation with a back end, certain
//   things may need to be checked further (e.g., isconnected)
void wmClearSocketsForThisTask( HTASK ht )
{
   PMYSOCKET pthissocket = pasocketshead ;
   int iSocketOffset ;

   for (iSocketOffset = 0; iSocketOffset < MAXSOCKETS ; iSocketOffset++)
   {
       // if this socket belongs to this task, make it available
       if ( ht == pthissocket[iSocketOffset].mytask )
       {
          wmClearSocketStruct( pthissocket ) ;
       }
   }
}

// no error checking, pointer assumed to come from wmGetSocketInfo()
void wmClearSocketStruct( PMYSOCKET pthissocket )
{
   if ( pthissocket )
   {
      if ( pthissocket->hRecv )
      {
         GlobalUnlock ( pthissocket->hRecv ) ;
         GlobalFree( pthissocket->hRecv ) ;
         pthissocket->hRecv = NULL ;
      }

      if ( pthissocket->hSend )
      {
         GlobalUnlock ( pthissocket->hSend ) ;
         GlobalFree( pthissocket->hSend ) ;
         pthissocket->hSend = NULL ;
      }

      pthissocket->mytask = NULL ;
      pthissocket->mys = SOCKET_ERROR ;
      pthissocket->isconnected = FALSE ;
      pthissocket->isstillreceiving = FALSE ;
      pthissocket->socketblocking = FALSE ;
      pthissocket->socketfamily = 0 ;
      pthissocket->socketprotocol = 0 ;
      pthissocket->sockettype = 0 ;
      pthissocket->wRecvSize = 0 ;
      pthissocket->wRecvOffset = 0 ;
   }
}

// loads a file from disk into a globally allocated block
//   the caller is responsible for unlocking and freeing this block
HANDLE wmLoadFile( LPSTR lpszFileName )
{
   HANDLE hMem ;
   HANDLE hFile ;
   int iLength ;
   LPSTR lpszBuffer ;

   hFile = _lopen( lpszFileName, OF_READ ) ;

   if ( -1 == hFile )
   {
      return( NULL ) ; // file not found
   }

   iLength = (int) _llseek( hFile, 0L, 2 ) ; // seek to end of file to get size
   if ( 0 == iLength )
   {
      _lclose( hFile ) ;
      return( NULL ) ; // file has no size
   }
   _llseek( hFile, 0L, 0 ) ; // restore file position to beginning

   hMem = GlobalAlloc( GMEM_MOVEABLE, iLength + 1 ) ; // leave room for null character
   if ( NULL == hMem )
   {
      return( NULL ) ; // allocation failed
   }

   lpszBuffer = GlobalLock( hMem ) ;

   // now read file into buffer
   _lread( hFile, lpszBuffer, iLength ) ;
   lpszBuffer[ iLength ] = '\0' ; // add null termination

   // if last physical char of file is EOF...
   if ( 0x1a == lpszBuffer[ --iLength ] )
   {
      lpszBuffer[ iLength ] = '\0' ; // ...replace EOF with NULL CHAR
   }

   GlobalUnlock( hMem ) ;
   _lclose( hFile ) ;

   return( hMem ) ;
}

